//
//  KBShader.m
//  CocoaShaders
//
//  Created by Guy English on 03/03/07.
//  Copyright 2007 __MyCompanyName__. All rights reserved.
//

#import "KBShader.h"

static NSMutableDictionary *namedShaders = nil;

// NULL shader - baseclass, draws nothing.
@implementation KBShader

- (void) encodeWithCoder: (NSCoder*) archiver
{
	NSString *name = [self name];
	if ( name ) [archiver encodeObject: name forKey: @"name"];
}

- (id) initWithCoder: (NSCoder*) unarchiver
{
	self = [self init];
	if ( !self ) return nil;
	
	NSString *name = [unarchiver decodeObjectForKey: @"name"];
	if ( name ) [self setName: name];
	
	return self;
}

- (void) drawInRect: (NSRect) rect input: (id) input
{
}

@end

// Conditional shader - will execute either the 'ifShader' or the 'elseShader' depending on the result
// of evaluating the predicate on the input
@implementation KBConditionalShader : KBShader

- (void) encodeWithCoder: (NSCoder*) archiver
{
	[super encodeWithCoder: archiver];
	[archiver encodeObject: _predicate forKey: @"predicate"];
	[archiver encodeObject: _ifShader forKey: @"ifShader"];
	[archiver encodeObject: _elseShader forKey: @"elseShader"];
}

- (id) initWithCoder: (NSCoder*) unarchiver
{
	self = [super initWithCoder: unarchiver];
	if ( !self ) return nil;
	
	_predicate = [[unarchiver decodeObjectForKey: @"predicate"] retain];
	_ifShader = [[unarchiver decodeObjectForKey: @"ifShader"] retain];
	_elseShader = [[unarchiver decodeObjectForKey: @"elseShader"] retain];
	
	return self;
}

- (id) initWithPredicate: (NSPredicate*) predicate ifShader: (KBShader*) ifShader elseShader: (KBShader*) elseShader
{
	self = [super init];
	if ( !self ) return nil;
	
	_predicate = [predicate retain];
	_ifShader = [ifShader retain];
	_elseShader = [elseShader retain];
	
	return self;
}

- (void) dealloc
{
	[_predicate release];
	[_ifShader release];
	[_elseShader release];
	[super dealloc];
}

- (void) drawInRect: (NSRect) rect input: (id) input
{
	if ( [_predicate evaluateWithObject: input] == YES )
	{
		[_ifShader drawInRect: rect input: input];
	}
	else
	{
		[_elseShader drawInRect: rect input: input];
	}
}

@end

// shader list - executes a list of shaders in sequence
@implementation KBShaderList : KBShader

- (void) encodeWithCoder: (NSCoder*) archiver
{
	[super encodeWithCoder: archiver];
	[archiver encodeObject: _shaders forKey: @"shaders"];
}

- (id) initWithCoder: (NSCoder*) unarchiver
{
	self = [super initWithCoder: unarchiver];
	if ( !self ) return nil;
	
	_shaders = [[unarchiver decodeObjectForKey: @"shaders"] retain];
	
	return self;
}

- (id) initWithShaders: (NSArray*) shaders
{
	self = [super init];
	if ( !self ) return nil;
	
	_shaders = [shaders copy];
	
	return self;
}

- (void) dealloc
{
	[_shaders release];
	[super dealloc];
}

- (void) drawInRect: (NSRect) rect input: (id) input
{
	unsigned i, max;
	max = [_shaders count];
	for ( i = 0; i < max; i++ )
	{
		[[_shaders objectAtIndex: i] drawInRect: rect input: input];
	}
}

@end

// clip shader - clips the given shader with the given bezier path
@implementation KBClipShader

- (void) encodeWithCoder: (NSCoder*) archiver
{
	[super encodeWithCoder: archiver];
	[archiver encodeObject: _clipPath forKey: @"archiver"];
	[archiver encodeObject: _shader forKey: @"shader"];
}

- (id) initWithCoder: (NSCoder*) unarchiver
{
	self = [super initWithCoder: unarchiver];
	if ( !self ) return nil;
	
	_clipPath = [[unarchiver decodeObjectForKey: @"archiver"] retain];
	_shader = [[unarchiver decodeObjectForKey: @"shader"] retain];
	
	return self;
}

- (id) initWithClipPath: (NSBezierPath*) path shader: (KBShader*) shader
{
	self = [super init];
	if ( !self ) return nil;
	
	_clipPath = [path retain];
	_shader = [shader retain];
	
	return self;
}

- (void) dealloc
{
	[_clipPath release];
	[_shader release];
	[super dealloc];
}

- (void) drawInRect: (NSRect) rect input: (id) input
{
    [NSGraphicsContext saveGraphicsState];
    [_clipPath addClip];
    [_shader drawInRect: rect input: input];
    [NSGraphicsContext restoreGraphicsState];
}

@end

// affine transform shader - transforms the drawing coord space while executing the given shader
@implementation KBAffineTransformShader

- (void) encodeWithCoder: (NSCoder*) archiver
{
	[super encodeWithCoder: archiver];
	[archiver encodeObject: _transform forKey: @"transform"];
	[archiver encodeObject: _shader forKey: @"shader"];
}

- (id) initWithCoder: (NSCoder*) unarchiver
{
	self = [super initWithCoder: unarchiver];
	if ( !self ) return nil;
	
	_transform = [[unarchiver decodeObjectForKey: @"transform"] retain];
	_shader = [[unarchiver decodeObjectForKey: @"shader"] retain];
	
	return self;
}

- (id) initWithAffineTransform: (NSAffineTransform*) transform shader: (KBShader*) shader
{
	self = [super init];
	if ( !self ) return nil;
	
	_transform = [transform retain];
	_shader = [shader retain];
	
	return self;
}

- (void) dealloc
{
	[_transform release];
	[_shader release];
	[super dealloc];
}

- (void) drawInRect: (NSRect) rect input: (id) input
{
	[NSGraphicsContext saveGraphicsState];
	[_transform concat];
	[_shader drawInRect: rect input: input];
	[NSGraphicsContext restoreGraphicsState];
}

@end

// fills an area with a solid color
@implementation KBSolidColorShader

- (void) encodeWithCoder: (NSCoder*) archiver
{
	[super encodeWithCoder: archiver];
	[archiver encodeObject: _color forKey: @"color"];
}

- (id) initWithCoder: (NSCoder*) unarchiver
{
	self = [super initWithCoder: unarchiver];
	if ( !self ) return nil;
	
	_color = [[unarchiver decodeObjectForKey: @"color"] retain];
	
	return self;
}

- (id) initWithColor: (NSColor*) color
{
	self = [super init];
	if ( !self ) return nil;
	
	_color = [color retain];
	
	return self;
}

- (void) dealloc
{
	[_color release];
	[super dealloc];
}

- (void) drawInRect: (NSRect) rect input: (id) input
{
	[_color set];
	NSRectFill( rect );
}

@end

// image shader - draws the image. can be set to various compositing modes and can control the drawing scale and source rect
@implementation KBImageShader

- (void) encodeWithCoder: (NSCoder*) archiver
{
	[super encodeWithCoder: archiver];
	[archiver encodeObject: _image forKey: @"image"];
	[archiver encodeRect: _srcRect forKey: @"srcRect"];
	[archiver encodeInt32: (int32_t)_operation forKey: @"operation"];
	[archiver encodeFloat: _fraction forKey: @"fraction"];
	[archiver encodeFloat: _scale forKey: @"scale"];
}

- (id) initWithCoder: (NSCoder*) unarchiver 
{
	self = [super initWithCoder: unarchiver];
	if ( !self ) return nil;
	
	_image = [[unarchiver decodeObjectForKey: @"image"] retain];
	_srcRect = [unarchiver decodeRectForKey: @"srcRect"];
	_operation = (NSCompositingOperation)[unarchiver decodeInt32ForKey: @"operation"];
	_fraction = [unarchiver decodeFloatForKey: @"fraction"];
	_scale = [unarchiver decodeFloatForKey: @"scale"];
	
	return self;
}

- (id) initWithImage: (NSImage*) image sourceRect: (NSRect) srcRect operation: (NSCompositingOperation) operation fraction: (float) fraction scale: (float) scale
{
	self = [super init];
	if ( !self ) return nil;
	
	_image = [image retain];
	_srcRect = srcRect;
	_operation = operation;
	_fraction = fraction;
	_scale = scale;
	
	return self;
}

- (id) initWithImage: (NSImage*) image
{
	return [self initWithImage: image sourceRect: NSMakeRect( 0, 0, [image size].width, [image size].height ) operation: NSCompositeSourceOver fraction: 1.0f scale: 1.0f];
}

- (void) dealloc
{
	[_image release];
	[super dealloc];
}

- (void) drawInRect: (NSRect) rect input: (id) input
{
	NSRect destRect = rect;
	destRect.size = [_image size];
	destRect.size.width *= _scale;
	destRect.size.height *= _scale;
	
	[_image drawInRect: destRect fromRect: _srcRect operation: _operation fraction: _fraction];
}

@end

// class and instance methods to deal with naming shader instances and retrieving them later.
@implementation KBShader ( KBShaderNaming )

+ (KBShader*) shaderNamed: (NSString*) name
{
	return [namedShaders objectForKey: name];
}

+ (void) setName: (NSString*) name forShader: (KBShader*) shader
{
	if ( namedShaders == nil )
	{
		namedShaders = [[NSMutableDictionary alloc] init];
	}
	
	[shader retain]; // retain here in case the only ref to us is in the namedShaders dictionary
	[namedShaders removeObjectsForKeys: [namedShaders allKeysForObject: shader]];
	[shader release];
	if ( name != nil )
	{
		[namedShaders setObject: shader forKey: name];
	}
}

- (void) setName: (NSString*) name
{
	[[self class] setName: name forShader: self];
}

- (NSString*) name
{
	NSArray *names = [namedShaders allKeysForObject: self];
	if ( [names count] == 0 ) return nil;
	return [names objectAtIndex: 0];
}

@end

// shader creation - handy class methods to return autoreleased shader instances.
@implementation KBShader ( KBShaderCreation )

+ (KBShader*) conditionalShaderWithPredicate: (NSPredicate*) predicate ifShader: (KBShader*) ifShader elseShader: (KBShader*) elseShader
{
	return [[[KBConditionalShader alloc] initWithPredicate: predicate ifShader: ifShader elseShader: elseShader] autorelease];
}

+ (KBShader*) shaderListWithShaders: (NSArray*) shaders
{
	return [[[KBShaderList alloc] initWithShaders: shaders] autorelease];
}

+ (KBShader*) clipShaderWithClipPath: (NSBezierPath*) path shader: (KBShader*) shader
{
	return [[[KBClipShader alloc] initWithClipPath: path shader: shader] autorelease];
}

+ (KBShader*) affineTransformShaderWithAffineTransform: (NSAffineTransform*) transform shader: (KBShader*) shader
{
	return [[[KBAffineTransformShader alloc] initWithAffineTransform: transform shader: shader] autorelease];
}

+ (KBShader*) imageShaderWithImage: (NSImage*) image
{
	return [[[KBImageShader alloc] initWithImage: image] autorelease];
}

+ (KBShader*) solidColorShaderWithColor: (NSColor*) color
{
	return [[[KBSolidColorShader alloc] initWithColor: color] autorelease];
}

@end

